@@ -15,6 +15,8 @@
 
 #include "soup-logger.h"
 #include "soup-message.h"
+#include "soup-session.h"
+#include "soup-session-feature.h"
 #include "soup-uri.h"
 
 /**
@@ -30,7 +32,7 @@
  * optionally configure it with soup_logger_set_request_filter(),
  * soup_logger_set_response_filter(), and soup_logger_set_printer(),
  * and then attach it to a session (or multiple sessions) with
- * soup_logger_attach().
+ * soup_session_add_feature().
  *
  * By default, the debugging output is sent to %stdout, and looks
  * something like:
@@ -64,9 +66,31 @@
  * across the lifetime of the #SoupLogger. In particular, this can be
  * used to identify when multiple messages are sent across the same
  * connection.
+ *
+ * The request half of the message is logged just before the first
+ * byte of the request gets written to the network (from the
+ * #SoupSession::request_started signal), and the response is logged
+ * just after the last byte of the response body is read from the
+ * network (from the #SoupMessage::got_body or
+ * #SoupMessage::got_informational signal). In particular, note that
+ * the #SoupMessage::got_headers signal, and anything triggered off it
+ * (such as #SoupSession::authenticate) will be emitted
+ * <emphasis>before</emphasis> the response headers are actually
+ * logged.
  **/
 
-G_DEFINE_TYPE (SoupLogger, soup_logger, G_TYPE_OBJECT)
+static void soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface, gpointer interface_data);
+
+static void request_queued  (SoupSessionFeature *feature, SoupSession *session,
+			     SoupMessage *msg);
+static void request_started  (SoupSessionFeature *feature, SoupSession *session,
+			      SoupMessage *msg, SoupSocket *socket);
+static void request_unqueued  (SoupSessionFeature *feature,
+			       SoupSession *session, SoupMessage *msg);
+
+G_DEFINE_TYPE_WITH_CODE (SoupLogger, soup_logger, G_TYPE_OBJECT,
+			 G_IMPLEMENT_INTERFACE (SOUP_TYPE_SESSION_FEATURE,
+						soup_logger_session_feature_init))
 
 typedef struct {
 	/* We use a mutex so that if requests are being run in
@@ -91,7 +115,6 @@ typedef struct {
 	SoupLoggerPrinter   printer;
 	gpointer            printer_data;
 	GDestroyNotify      printer_dnotify;
-	char                direction;
 } SoupLoggerPrivate;
 #define SOUP_LOGGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_LOGGER, SoupLoggerPrivate))
 
@@ -134,6 +157,15 @@ soup_logger_class_init (SoupLoggerClass *logger_class)
 	object_class->finalize = finalize;
 }
 
+static void
+soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface,
+				  gpointer interface_data)
+{
+	feature_interface->request_queued = request_queued;
+	feature_interface->request_started = request_started;
+	feature_interface->request_unqueued = request_unqueued;
+}
+
 /**
  * SoupLoggerLogLevel:
  * @SOUP_LOGGER_LOG_NONE: No logging
@@ -158,6 +190,8 @@ soup_logger_class_init (SoupLoggerClass *logger_class)
  * If you need finer control over what message parts are and aren't
  * logged, use soup_logger_set_request_filter() and
  * soup_logger_set_response_filter().
+ *
+ * Returns: a new #SoupLogger
  **/
 SoupLogger *
 soup_logger_new (SoupLoggerLogLevel level, int max_body_size) 
@@ -312,23 +346,6 @@ soup_logger_set_id (SoupLogger *logger, gpointer object)
 	return GPOINTER_TO_UINT (id);
 }
 
-static void
-soup_logger_clear_id (SoupLogger *logger, gpointer object)
-{
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
-
-	g_object_set_qdata (object, priv->tag, NULL);
-}
-
-static void request_started (SoupSession *session, SoupMessage *msg,
-			     SoupSocket *socket, gpointer user_data);
-
-static void
-weak_notify_unref (gpointer logger, GObject *ex_session)
-{
-	g_object_unref (logger);
-}
-
 /**
  * soup_logger_attach:
  * @logger: a #SoupLogger
@@ -340,18 +357,14 @@ weak_notify_unref (gpointer logger, GObject *ex_session)
  * (The session will take a reference on @logger, which will be
  * removed when you call soup_logger_detach(), or when the session is
  * destroyed.)
+ *
+ * Deprecated: Use soup_session_add_feature() instead.
  **/
 void
 soup_logger_attach (SoupLogger  *logger,
 		    SoupSession *session)
 {
-	if (!soup_logger_get_id (logger, session))
-		soup_logger_set_id (logger, session);
-	g_signal_connect (session, "request_started",
-			  G_CALLBACK (request_started), logger);
-
-	g_object_weak_ref (G_OBJECT (session),
-			   weak_notify_unref, g_object_ref (logger));
+	soup_session_add_feature (session, SOUP_SESSION_FEATURE (logger));
 }
 
 /**
@@ -360,20 +373,19 @@ soup_logger_attach (SoupLogger  *logger,
  * @session: a #SoupSession
  *
  * Stops @logger from watching @session.
+ *
+ * Deprecated: Use soup_session_remove_feature() instead.
  **/
 void
 soup_logger_detach (SoupLogger  *logger,
 		    SoupSession *session)
 {
-	g_signal_handlers_disconnect_by_func (session, request_started, logger);
-
-	g_object_weak_unref (G_OBJECT (session),
-			     weak_notify_unref, logger);
+	soup_session_remove_feature (session, SOUP_SESSION_FEATURE (logger));
 }
 
 static void
 soup_logger_print (SoupLogger *logger, SoupLoggerLogLevel level,
-		   const char *format, ...)
+		   char direction, const char *format, ...)
 {
 	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
 	va_list args;
@@ -394,10 +406,10 @@ soup_logger_print (SoupLogger *logger, SoupLoggerLogLevel level,
 		if (end)
 			*end = '\0';
 		if (priv->printer) {
-			priv->printer (logger, level, priv->direction,
+			priv->printer (logger, level, direction,
 				       line, priv->printer_data);
 		} else
-			printf ("%c %s\n", priv->direction, line);
+			printf ("%c %s\n", direction, line);
 
 		line = end + 1;
 	} while (end && *line);
@@ -406,28 +418,22 @@ soup_logger_print (SoupLogger *logger, SoupLoggerLogLevel level,
 }
 
 static void
-print_header (const char *name, const char *value, gpointer logger)
+soup_logger_print_basic_auth (SoupLogger *logger, const char *value)
 {
-	if (!g_ascii_strcasecmp (name, "Authorization") &&
-	    !g_ascii_strncasecmp (value, "Basic ", 6)) {
-		char *decoded, *p;
-		gsize len;
+	char *decoded, *p;
+	gsize len;
 
-		decoded = (char *)g_base64_decode (value + 6, &len);
-		if (!decoded)
-			decoded = g_strdup (value);
-		p = strchr (decoded, ':');
-		if (p) {
-			while (++p < decoded + len)
-				*p = '*';
-		}
-		soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS,
-				   "%s: Basic [%.*s]", name, len, decoded);
-		g_free (decoded);
-	} else {
-		soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS,
-				   "%s: %s", name, value);
+	decoded = (char *)g_base64_decode (value + 6, &len);
+	if (!decoded)
+		decoded = g_strdup (value);
+	p = strchr (decoded, ':');
+	if (p) {
+		while (++p < decoded + len)
+			*p = '*';
 	}
+	soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS, '>',
+			   "Authorization: Basic [%.*s]", len, decoded);
+	g_free (decoded);
 }
 
 static void
@@ -437,6 +443,8 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 {
 	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
 	SoupLoggerLogLevel log_level;
+	SoupMessageHeadersIter iter;
+	const char *name, *value;
 	SoupURI *uri;
 
 	if (priv->request_filter) {
@@ -448,16 +456,14 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 	if (log_level == SOUP_LOGGER_LOG_NONE)
 		return;
 
-	priv->direction = '>';
-
 	uri = soup_message_get_uri (msg);
 	if (msg->method == SOUP_METHOD_CONNECT) {
-		soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL,
+		soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '>',
 				   "CONNECT %s:%u HTTP/1.%d",
 				   uri->host, uri->port,
 				   soup_message_get_http_version (msg));
 	} else {
-		soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL,
+		soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '>',
 				   "%s %s%s%s HTTP/1.%d",
 				   msg->method, uri->path,
 				   uri->query ? "?" : "",
@@ -465,10 +471,10 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 				   soup_message_get_http_version (msg));
 	}
 
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL,
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '>',
 			   "Soup-Debug-Timestamp: %lu",
 			   (unsigned long)time (0));
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL,
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '>',
 			   "Soup-Debug: %s %u (%p), %s %u (%p), %s %u (%p)%s",
 			   g_type_name_from_instance ((GTypeInstance *)session),
 			   soup_logger_get_id (logger, session), session,
@@ -481,9 +487,22 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 	if (log_level == SOUP_LOGGER_LOG_MINIMAL)
 		return;
 
-	print_header ("Host", uri->host, logger);
-	soup_message_headers_foreach (msg->request_headers,
-				      print_header, logger);
+	if (!soup_message_headers_get (msg->request_headers, "Host")) {
+		soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS, '>',
+				   "Host: %s%c%u", uri->host,
+				   soup_uri_uses_default_port (uri) ? '\0' : ':',
+				   uri->port);
+	}
+	soup_message_headers_iter_init (&iter, msg->request_headers);
+	while (soup_message_headers_iter_next (&iter, &name, &value)) {
+		if (!g_ascii_strcasecmp (name, "Authorization") &&
+		    !g_ascii_strncasecmp (value, "Basic ", 6))
+			soup_logger_print_basic_auth (logger, value);
+		else {
+			soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS, '>',
+					   "%s: %s", name, value);
+		}
+	}
 	if (log_level == SOUP_LOGGER_LOG_HEADERS)
 		return;
 
@@ -494,7 +513,7 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 		soup_buffer_free (request);
 
 		if (soup_message_headers_get_expectations (msg->request_headers) != SOUP_EXPECTATION_CONTINUE) {
-			soup_logger_print (logger, SOUP_LOGGER_LOG_BODY,
+			soup_logger_print (logger, SOUP_LOGGER_LOG_BODY, '>',
 					   "\n%s", msg->request_body->data);
 		}
 	}
@@ -505,6 +524,8 @@ print_response (SoupLogger *logger, SoupMessage *msg)
 {
 	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
 	SoupLoggerLogLevel log_level;
+	SoupMessageHeadersIter iter;
+	const char *name, *value;
 
 	if (priv->response_filter) {
 		log_level = priv->response_filter (logger, msg,
@@ -515,17 +536,15 @@ print_response (SoupLogger *logger, SoupMessage *msg)
 	if (log_level == SOUP_LOGGER_LOG_NONE)
 		return;
 
-	priv->direction = '<';
-
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL,
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '<',
 			   "HTTP/1.%d %u %s\n",
 			   soup_message_get_http_version (msg),
 			   msg->status_code, msg->reason_phrase);
 
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL,
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '<',
 			   "Soup-Debug-Timestamp: %lu",
 			   (unsigned long)time (0));
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL,
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '<',
 			   "Soup-Debug: %s %u (%p)",
 			   g_type_name_from_instance ((GTypeInstance *)msg),
 			   soup_logger_get_id (logger, msg), msg);
@@ -533,13 +552,16 @@ print_response (SoupLogger *logger, SoupMessage *msg)
 	if (log_level == SOUP_LOGGER_LOG_MINIMAL)
 		return;
 
-	soup_message_headers_foreach (msg->response_headers,
-				     print_header, logger);
+	soup_message_headers_iter_init (&iter, msg->response_headers);
+	while (soup_message_headers_iter_next (&iter, &name, &value)) {
+		soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS, '<',
+				   "%s: %s", name, value);
+	}
 	if (log_level == SOUP_LOGGER_LOG_HEADERS)
 		return;
 
 	if (msg->response_body->length) {
-		soup_logger_print (logger, SOUP_LOGGER_LOG_BODY,
+		soup_logger_print (logger, SOUP_LOGGER_LOG_BODY, '<',
 				   "\n%s", msg->response_body->data);
 	}
 }
@@ -553,14 +575,12 @@ got_informational (SoupMessage *msg, gpointer user_data)
 	g_mutex_lock (priv->lock);
 
 	print_response (logger, msg);
-	priv->direction = ' ';
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, "");
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
 
 	if (msg->status_code == SOUP_STATUS_CONTINUE && msg->request_body->data) {
 		SoupLoggerLogLevel log_level;
 
-		priv->direction = '>';
-		soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL,
+		soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '>',
 				   "[Now sending request body...]");
 
 		if (priv->request_filter) {
@@ -570,12 +590,11 @@ got_informational (SoupMessage *msg, gpointer user_data)
 			log_level = priv->level;
 
 		if (log_level == SOUP_LOGGER_LOG_BODY) {
-			soup_logger_print (logger, SOUP_LOGGER_LOG_BODY,
+			soup_logger_print (logger, SOUP_LOGGER_LOG_BODY, '>',
 					   "%s", msg->request_body->data);
 		}
 
-		priv->direction = ' ';
-		soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, "");
+		soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
 	}
 
 	g_mutex_unlock (priv->lock);
@@ -590,30 +609,28 @@ got_body (SoupMessage *msg, gpointer user_data)
 	g_mutex_lock (priv->lock);
 
 	print_response (logger, msg);
-	priv->direction = ' ';
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, "");
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
 
 	g_mutex_unlock (priv->lock);
 }
 
 static void
-finished_handler (SoupMessage *msg, gpointer user_data)
+request_queued (SoupSessionFeature *logger, SoupSession *session,
+		SoupMessage *msg)
 {
-	SoupLogger *logger = user_data;
-
-	g_signal_handlers_disconnect_by_func (msg, got_informational, logger);
-	g_signal_handlers_disconnect_by_func (msg, got_body, logger);
-	g_signal_handlers_disconnect_by_func (msg, finished_handler, logger);
-
-	soup_logger_clear_id (logger, msg);
+	g_signal_connect (msg, "got-informational",
+			  G_CALLBACK (got_informational),
+			  logger);
+	g_signal_connect (msg, "got-body",
+			  G_CALLBACK (got_body),
+			  logger);
 }
 
 static void
-request_started (SoupSession *session, SoupMessage *msg,
-		 SoupSocket *socket, gpointer user_data)
+request_started (SoupSessionFeature *feature, SoupSession *session,
+		 SoupMessage *msg, SoupSocket *socket)
 {
-	SoupLogger *logger = user_data;
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLogger *logger = SOUP_LOGGER (feature);
 	gboolean restarted;
 	guint msg_id;
 
@@ -621,24 +638,24 @@ request_started (SoupSession *session, SoupMessage *msg,
 	if (msg_id)
 		restarted = TRUE;
 	else {
-		msg_id = soup_logger_set_id (logger, msg);
+		soup_logger_set_id (logger, msg);
 		restarted = FALSE;
-
-		g_signal_connect (msg, "got-informational",
-				  G_CALLBACK (got_informational),
-				  logger);
-		g_signal_connect (msg, "got-body",
-				  G_CALLBACK (got_body),
-				  logger);
-		g_signal_connect (msg, "finished",
-				  G_CALLBACK (finished_handler),
-				  logger);
 	}
 
+	if (!soup_logger_get_id (logger, session))
+		soup_logger_set_id (logger, session);
+
 	if (!soup_logger_get_id (logger, socket))
 		soup_logger_set_id (logger, socket);
 
 	print_request (logger, msg, session, socket, restarted);
-	priv->direction = ' ';
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, "");
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
+}
+
+static void
+request_unqueued (SoupSessionFeature *logger, SoupSession *session,
+		  SoupMessage *msg)
+{
+	g_signal_handlers_disconnect_by_func (msg, got_informational, logger);
+	g_signal_handlers_disconnect_by_func (msg, got_body, logger);
 }
